#ifndef MATH3D_PRIMITIVE_TESTS_HPP
#define MATH3D_PRIMITIVE_TESTS_HPP

/**
 *	@file	
 *	@brief	Primitive collision test functions.
 *	@author	Tomohiro Matsumoto
 */

#include <gpuppur/utility/begin_suppress_warnings_from_others_code.hpp>
#include <boost/call_traits.hpp>
#include <gpuppur/utility/end_suppress_warnings.hpp>

#include <gpuppur/3dmath/vectorNd.hpp>

namespace gpuppur
{

template<typename Vector, typename AABB>
bool test_aabb_triangle
(
	const Vector& v0,
	const Vector& v1,
	const Vector& v2,
	const AABB&		aabb
)
{
	//Using Separating-axis test.

	typedef Vector								vector_type;
	typedef typename Vector::value_type			value_type;
	typedef gpuppur::VectorNd<vector_type, 3>	triangle_type;

	vector_type center = (aabb.get_max() + aabb.get_min())/2;
	vector_type extents = aabb.get_size()/2;

	triangle_type tr(v0 - center, v1 - center, v2 - center); 
	triangle_type edges(tr[1] - tr[0], tr[2] - tr[1], tr[0] - tr[2]);

	//refer real time collision detection page 171
/*	I'll write fast expanded code someday....
	p0 = -tr[0][1]*tr[1][2] + tr[0][2]*tr[1][1];
	p2 = tr[2][2]*(tr[1][1] - tr[0][1]) - tr[2][2]*(tr[1][2] - tr[0][2]);
*/

	for(std::size_t i=0; i<Vector::dimention; ++i)
	{
		vector_type aabb_face_normal(0);
		aabb_face_normal[i] = 1;

		for(std::size_t j=0; j<Vector::dimention; ++j)
		{
			vector_type separating_axis(aabb_face_normal.outerProduct(edges[j]));

			value_type aabb_rad = 0;
			for(std::size_t k=0; k<Vector::dimention; ++k)
			{
				aabb_rad += abs(extents[k]*separating_axis[k]);
			}

			vector_type proj_tri;
			for(std::size_t k=0; k<Vector::dimention; ++k)
			{
				proj_tri[k] = tr[k].innerProduct(separating_axis);
			}

			if
			(
				std::max
				(
					-std::max(std::max(proj_tri[0], proj_tri[1]), proj_tri[2]), 
					std::min(std::min(proj_tri[0], proj_tri[1]), proj_tri[2])
				)
				> aabb_rad
			)
			{
				return false;
			}
		}
	}

	for(std::size_t i=0; i<Vector::dimention; ++i)
	{
		if
		(
			std::max(std::max(tr[0][i], tr[1][i]), tr[2][i]) < -abs(extents[i])
			||
			std::min(std::min(tr[0][i], tr[1][i]), tr[2][i]) > abs(extents[i])
		)
		{
			return false;
		}
	}

	vector_type normal = edges[0].outerProduct(-edges[2]).getNormalized();

	value_type d = normal.innerProduct(tr[0]);

	value_type aabb_rad = 0;
	for(std::size_t k=0; k<Vector::dimention; ++k)
	{
		aabb_rad += abs(extents[k]*normal[k]);
	}

	return abs(d) <= aabb_rad;
}

}	// end of namespace gpuppur

#endif
