#ifndef GPUPPUR_CPU_MESH_HPP
#define GPUPPUR_CPU_MESH_HPP

/**
 *	@file	
 *	@brief	Mesh class of CPU Raytracer class
 *	@author	Tomohiro Matsumoto
 */

#include <gpuppur/utility/begin_suppress_warnings_from_others_code.hpp>
#include <vector>
#include <limits>
#include <list>
#include <boost/array.hpp>
#include <boost/mpl/if.hpp>
#include <boost/mpl/int.hpp>
#include <gpuppur/utility/end_suppress_warnings.hpp>

#include <gpuppur/3dmath/triangle.hpp>
#include <gpuppur/3dmath/axis_aligned_bounding_box.hpp>

#include "mesh_accelerators/octree.hpp"

namespace gpuppur
{
namespace cpu_raytracers
{
namespace detail
{

class mesh : public
			octree_mesh<>
{
public:

	typedef
	octree_mesh<>
	base;
	typedef mesh							this_type;
	typedef std::list<this_type>::iterator
											iterator;
/*
	mesh_main()
	{
	}
*/
	mesh
	(
		const std::vector<vector3>&			vertices,
		const std::vector<unsigned short>&	triangles
	):
		base(vertices, triangles)
	{
	}

	mesh
	(
		const std::vector<vector3>&			vertices,
		const std::vector<unsigned int>&	triangles
	):
		base(vertices, triangles)
	{
	}

	void set_self_iterator(iterator itr)
	{
		this->self = itr;
	}

	const iterator& get_self_iterator() const
	{
		return this->self;
	}

	
	template<class Parent, class Base>
	class handler_to_mesh : public Base
	{
	public:

		typedef Base						base;
		typedef this_type					handled_type;
		typedef gpuppur::mesh_virtual		virtual_type;
		typedef gpuppur::mesh_generic		generic_type;
		typedef gpuppur::cpu_raytracers::detail::cpu_raytracer_implement
											friend_type;
	public:

		handled_type& get_handled()
		{
			return reinterpret_cast<Parent*>(this)->get_handle();
		}

		const handled_type& get_handled() const
		{
			return reinterpret_cast<const Parent*>(this)->get_handle();
		}

	protected:

		friend void release_mesh(handled_type* mesh);

		handler_to_mesh()
		{
		}

		static void release(handled_type* mesh)
		{
			gpuppur::cpu_raytracers::detail::release_mesh(mesh);
		}
	};

private:

	iterator		self;		//!Iterator to std::list<this class> for fast erase object from pointer to this class.
};


}	// end of namespace detail
}	// end of namespace cpu_raytracers
}	// end of namespace gpuppur

#endif

