#ifndef GPUPPUR_CPU_MESH_ACCELERATORS_SIMPLE_MESH_HPP
#define GPUPPUR_CPU_MESH_ACCELERATORS_SIMPLE_MESH_HPP

/**
 *	@file	
 *	@brief	Base Mesh class of CPU Raytracer class
 *	@author	Tomohiro Matsumoto
 */

#include <gpuppur/utility/begin_suppress_warnings_from_others_code.hpp>
#include <vector>
#include <boost/mpl/int.hpp>
#include <boost/tuple/tuple.hpp>
#include <gpuppur/utility/end_suppress_warnings.hpp>

#include <gpuppur/3dmath/vectorNd.hpp>
#include <gpuppur/3dmath/axis_aligned_bounding_box.hpp>

#include "base_mesh.hpp"

namespace gpuppur
{
namespace cpu_raytracers
{
namespace detail
{

template<template<class> class Base = base_mesh<>::template c>
class simple_mesh : public Base<simple_mesh<Base> >
{
public:

	typedef Base<simple_mesh<Base> >	base;
	typedef simple_mesh					this_type;

private:

	template<int BytesPerIndex>
	bool cast_ray_impl
	(
		const ray& casted_ray,
		float max_t,
		vector3& pos,
		vector3& normal,
		boost::mpl::int_<BytesPerIndex>
	) const
	{
		if(!base::cast_ray_to_aabb(casted_ray, max_t))
		{
			return false;
		}

		float min_t = std::numeric_limits<float>::max();
		vector3 tmp_pos, tmp_normal;
		std::size_t i=0,
					num_indices
					=
					base::get_num_indices(boost::mpl::int_<BytesPerIndex>());

		for(i=0; i<num_indices; i+=3)
		{
			float t =
			gpuppur::triangle<vector3>::get_cross
			(
				casted_ray.getPos(),
				casted_ray.getDir(),
				base::get_vertex_from_index_of_index
				(
					boost::mpl::int_<BytesPerIndex>(), i+0
				),
				base::get_vertex_from_index_of_index
				(
					boost::mpl::int_<BytesPerIndex>(), i+1
				),
				base::get_vertex_from_index_of_index
				(
					boost::mpl::int_<BytesPerIndex>(), i+2
				),
				tmp_pos, tmp_normal
			);

			if(t < 0.0f || t > max_t || t >= min_t)
			{
				continue;
			}

			min_t = t;
			pos = tmp_pos;
			normal = tmp_normal;
		}

		return min_t <= max_t ? true : false;
	}


public:

	simple_mesh()
	{
	}

	simple_mesh
	(
		const std::vector<vector3>&			vertices,
		const std::vector<unsigned short>&	triangles
	):
		base(vertices, triangles)
	{
	}

	simple_mesh
	(
		const std::vector<vector3>&			vertices,
		const std::vector<unsigned int>&	triangles
	):
		base(vertices, triangles)
	{
	}

	void analyze_data()
	{
		PRINT_VAL(this->get_num_indices()/3);
	}

	friend class base;
};

}	// end of namespace detail
}	// end of namespace cpu_raytracers
}	// end of namespace gpuppur

#endif

